/* Driver for Philips webcam
   Functions that send various control messages to the webcam, including
   video modes.
   (C) 1999-2002 Nemosoft Unv. (webcam@smcc.demon.nl)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
   Changes
   2001/08/03  Alvarado   Added methods for changing white balance and 
                          red/green gains
 */

/* Control functions for the cam; brightness, contrast, video mode, etc. */


/*
 Changes:

  This is now part of the wctweak_lib project to build a static Win32 library

  Additions and changes and mostly deletions for a Win32 Version by BM 20031112
  This is only to support the IOCTLs for the cameras to fiddle  with some features 
  not available in the regular driver all the video stuff has been deleted.

  USB communication takes place through the libusb-win32 drivers

  All restrictions and applications mentioned above still apply to this file
   -- wctweak_lib project (C)2003 M. Burri 

  adapted for the libusb-win32 calling interface
  have deleted most of the code - otherwise it was to crowded with things not to 
  bother with....

  I've renamed the public interface from pwc_ to wct_ as it does not provide
  the full wet of pwc 8.6 and shoul not be mixed up with it.

  Thank you guys for the great work done.

*/

#include <errno.h>
#include <string.h>
 
#include <include\usb.h>  //BM from libusb-win32 project

#include "wct.h"
//#include "wct-ioctl.h"


/* Request types: video */
#define SET_LUM_CTL			0x01
#define GET_LUM_CTL			0x02
#define SET_CHROM_CTL			0x03
#define GET_CHROM_CTL			0x04
#define SET_STATUS_CTL			0x05
#define GET_STATUS_CTL			0x06
#define SET_EP_STREAM_CTL		0x07
#define GET_EP_STREAM_CTL		0x08
#define SET_MPT_CTL			0x0D
#define GET_MPT_CTL			0x0E

/* Selectors for the Luminance controls [GS]ET_LUM_CTL */
#define AGC_MODE_FORMATTER			0x2000
#define PRESET_AGC_FORMATTER			0x2100
#define SHUTTER_MODE_FORMATTER			0x2200
#define PRESET_SHUTTER_FORMATTER		0x2300
#define PRESET_CONTOUR_FORMATTER		0x2400
#define AUTO_CONTOUR_FORMATTER			0x2500
#define BACK_LIGHT_COMPENSATION_FORMATTER	0x2600
#define CONTRAST_FORMATTER			0x2700
#define DYNAMIC_NOISE_CONTROL_FORMATTER		0x2800
#define FLICKERLESS_MODE_FORMATTER		0x2900
#define AE_CONTROL_SPEED			0x2A00
#define BRIGHTNESS_FORMATTER			0x2B00
#define GAMMA_FORMATTER				0x2C00

/* Selectors for the Chrominance controls [GS]ET_CHROM_CTL */
#define WB_MODE_FORMATTER			0x1000
#define AWB_CONTROL_SPEED_FORMATTER		0x1100
#define AWB_CONTROL_DELAY_FORMATTER		0x1200
#define PRESET_MANUAL_RED_GAIN_FORMATTER	0x1300
#define PRESET_MANUAL_BLUE_GAIN_FORMATTER	0x1400
#define COLOUR_MODE_FORMATTER			0x1500
#define SATURATION_MODE_FORMATTER1		0x1600
#define SATURATION_MODE_FORMATTER2		0x1700

/* Selectors for the Status controls [GS]ET_STATUS_CTL */
#define SAVE_USER_DEFAULTS_FORMATTER		0x0200
#define RESTORE_USER_DEFAULTS_FORMATTER		0x0300
#define RESTORE_FACTORY_DEFAULTS_FORMATTER	0x0400
#define READ_AGC_FORMATTER			0x0500
#define READ_SHUTTER_FORMATTER			0x0600
#define READ_RED_GAIN_FORMATTER			0x0700
#define READ_BLUE_GAIN_FORMATTER		0x0800
#define SENSOR_TYPE_FORMATTER1			0x0C00
#define READ_RAW_Y_MEAN_FORMATTER		0x3100
#define SET_POWER_SAVE_MODE_FORMATTER		0x3200
#define MIRROR_IMAGE_FORMATTER			0x3300
#define LED_FORMATTER				0x3400
#define SENSOR_TYPE_FORMATTER2			0x3700

/* Formatters for the Video Endpoint controls [GS]ET_EP_STREAM_CTL */
#define VIDEO_OUTPUT_CONTROL_FORMATTER		0x0100

/* Formatters for the motorized pan & tilt [GS]ET_MPT_CTL */
#define PT_RELATIVE_CONTROL_FORMATTER		0x01
#define PT_RESET_CONTROL_FORMATTER		0x02
#define PT_STATUS_FORMATTER			0x03

#define HZ 500  // BM timeout ?!

static char *size2name[PSZ_MAX] =
{
	"subQCIF",
	"QSIF",
	"QCIF",
	"SIF",
	"CIF",
	"VGA",
};  


/********/


/* Entries for the Nala (645/646) camera; the Nala doesn't have compression 
   preferences, so you either get compressed or non-compressed streams.
   
   An alternate value of 0 means this mode is not available at all.
 */

struct Nala_table_entry {
	char alternate;			/* USB alternate setting */
	int compressed;			/* Compressed yes/no */

	unsigned char mode[3];		/* precomputed mode table */
};

static struct Nala_table_entry Nala_table[PSZ_MAX][8] =
{
#include "wct_nala.h"
};

/* This tables contains entries for the 675/680/690 (Timon) camera, with
   4 different qualities (no compression, low, medium, high).
   It lists the bandwidth requirements for said mode by its alternate interface 
   number. An alternate of 0 means that the mode is unavailable.
   
   There are 6 * 4 * 4 entries: 
     6 different resolutions subqcif, qsif, qcif, sif, cif, vga
     6 framerates: 5, 10, 15, 20, 25, 30
     4 compression modi: none, low, medium, high
     
   When an uncompressed mode is not available, the next available compressed mode 
   will be chosen (unless the decompressor is absent). Sometimes there are only
   1 or 2 compressed modes available; in that case entries are duplicated.
*/
struct Timon_table_entry 
{
	char alternate;			/* USB alternate interface */
	unsigned short packetsize;	/* Normal packet size */
	unsigned short bandlength;	/* Bandlength when decompressing */
	unsigned char mode[13];		/* precomputed mode settings for cam */
};

static struct Timon_table_entry Timon_table[PSZ_MAX][6][4] = 
{
#include "wct_timon.h"
};

/* Entries for the Kiara (730/740/750) camera */

struct Kiara_table_entry
{
	char alternate;			/* USB alternate interface */
	unsigned short packetsize;	/* Normal packet size */
	unsigned short bandlength;	/* Bandlength when decompressing */
	unsigned char mode[12];		/* precomputed mode settings for cam */
};

static struct Kiara_table_entry Kiara_table[PSZ_MAX][6][4] =
{
#include "wct_kiara.h"
};


/****************************************************************************/


#define SendControlMsg(request, value, buflen) \
	usb_control_msg(pdev->udevH,  \
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE, \
		request, \
		value, \
		pdev->vcinterface, \
		&buf, buflen, HZ / 2)

#define RecvControlMsg(request, value, buflen) \
	usb_control_msg(pdev->udevH,  \
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE, \
		request, \
		value, \
		pdev->vcinterface, \
		&buf, buflen, HZ / 2)


#if wct_DEBUG
void wct_hexdump(void *p, int len)
{
	int i;
	unsigned char *s;
	char buf[100], *d;
	
	s = (unsigned char *)p;
	d = buf;
	*d = '\0';
	Debug("Doing hexdump @ %p, %d bytes.\n", p, len);
	for (i = 0; i < len; i++) {
		d += sprintf(d, "%02X ", *s++);
		if ((i & 0xF) == 0xF) {
			Debug("%s\n", buf);
			d = buf;
			*d = '\0';
		}
	}
	if ((i & 0xF) != 0)
		Debug("%s\n", buf);
}
#endif


int send_video_command(struct usb_dev_handle *udev, int index, void *buf, int buflen)
{
	return usb_control_msg(udev,
		//usb_sndctrlpipe(udev, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_EP_STREAM_CTL,
		VIDEO_OUTPUT_CONTROL_FORMATTER,
		index,
		buf, buflen, HZ);
}



int set_video_mode_Nala(struct wct_device *pdev, int size, int frames)
{
	unsigned char buf[3];
	int ret, fps;
	struct Nala_table_entry *pEntry;
	int frames2frames[31] = 
	{ /* closest match of framerate */
	   0,  0,  0,  0,  4,  /*  0-4  */
	   5,  5,  7,  7, 10,  /*  5-9  */
          10, 10, 12, 12, 15,  /* 10-14 */
          15, 15, 15, 20, 20,  /* 15-19 */
          20, 20, 20, 24, 24,  /* 20-24 */
          24, 24, 24, 24, 24,  /* 25-29 */
          24                   /* 30    */
	};
	int frames2table[31] = 
	{ 0, 0, 0, 0, 0, /*  0-4  */
	  1, 1, 1, 2, 2, /*  5-9  */
	  3, 3, 4, 4, 4, /* 10-14 */
	  5, 5, 5, 5, 5, /* 15-19 */
	  6, 6, 6, 6, 7, /* 20-24 */
	  7, 7, 7, 7, 7, /* 25-29 */
	  7              /* 30    */
	};
	
	if (size < 0 || size > PSZ_CIF || frames < 4 || frames > 25)
		return -EINVAL;
	frames = frames2frames[frames];
	fps = frames2table[frames];
	pEntry = &Nala_table[size][fps];
	if (pEntry->alternate == 0)
		return -EINVAL;

	memcpy(buf, pEntry->mode, 3);	
	ret = send_video_command(pdev->udevH, pdev->vendpoint, buf, 3);
	if (ret < 0)
		return ret;
	return 0;
}


int set_video_mode_Timon(struct wct_device *pdev, int size, int frames, int compression, int snapshot)
{
	unsigned char buf[13];
	struct Timon_table_entry *pChoose;
	int ret, fps;

	if (size >= PSZ_MAX || frames < 5 || frames > 30 || compression < 0 || compression > 3)
		return -EINVAL;
	if (size == PSZ_VGA && frames > 15)
		return -EINVAL;
	fps = (frames / 5) - 1;
	
	/* Find a supported framerate with progressively higher compression ratios
	   if the preferred ratio is not available.
	*/
	pChoose = NULL;
	while (compression <= 3) {
		pChoose = &Timon_table[size][fps][compression];
		if (pChoose->alternate != 0)
			break;
		compression++;	
	}
	if (pChoose == NULL || pChoose->alternate == 0)
		return -ENOENT; /* Not supported. */

	memcpy(buf, pChoose->mode, 13);
	if (snapshot)
		buf[0] |= 0x80;
	ret = send_video_command(pdev->udevH, pdev->vendpoint, buf, 13);
	if (ret < 0)
		return ret;
	return 0;
}


int set_video_mode_Kiara(struct wct_device *pdev, int size, int frames, int compression, int snapshot)
{
	struct Kiara_table_entry *pChoose;
	int fps, ret;
	unsigned char buf[12];
	
	if (size >= PSZ_MAX || frames < 5 || frames > 30 || compression < 0 || compression > 3)
		return -EINVAL;
	if (size == PSZ_VGA && frames > 15)
		return -EINVAL;
	fps = (frames / 5) - 1;
	
	/* Find a supported framerate with progressively higher compression ratios
	   if the preferred ratio is not available.
	*/
	pChoose = NULL;
	while (compression <= 3) {
		pChoose = &Kiara_table[size][fps][compression];
		if (pChoose->alternate != 0)
			break;
		compression++;	
	}
	if (pChoose == NULL || pChoose->alternate == 0)
		return -ENOENT; /* Not supported. */

	/* usb_control_msg won't take staticly allocated arrays as argument?? */
	memcpy(buf, pChoose->mode, 12);
	if (snapshot)
		buf[0] |= 0x80;

	/* Firmware bug: video endpoint is 5, but commands are sent to endpoint 4 */
	ret = send_video_command(pdev->udevH, 4 /* pdev->vendpoint */, buf, 12);
	if (ret < 0)
		return ret;
	return 0;
}


struct wct_coord
{
	int x, y;		/* guess what */
	int size;		/* size, or offset */
};

struct wct_coord wct_image_sizes[PSZ_MAX] = 
{
	{ 128,  96, 0 },
	{ 160, 120, 0 },
	{ 176, 144, 0 },
	{ 320, 240, 0 },
	{ 352, 288, 0 },
	{ 640, 480, 0 },
};

/* x,y -> PSZ_ */
int wct_decode_size(struct wct_device *pdev, int width, int height)
{
	int i, find;

	/* Make sure we don't go beyond our max size */
//	if (width > pdev->view_max.x || height > pdev->view_max.y)
//		return -1;
	/* Find the largest size supported by the camera that fits into the
	   requested size. 
	 */
	find = -1;
	for (i = 0; i < PSZ_MAX; i++)
		if (wct_image_sizes[i].x <= width && wct_image_sizes[i].y <= height)
			find = i;

	return find;
}

/**
   @pdev: device structure
   @width: viewport width
   @height: viewport height
   @frame: framerate, in fps
   @compression: preferred compression ratio
   @snapshot: snapshot mode or streaming
 */
int wct_set_video_mode(struct wct_device *pdev, int width, int height, int frames, int compression, int snapshot)
{
	int ret, size;
	
	size = wct_decode_size(pdev, width, height);
	if (size < 0) {
//		printf("Could not find suitable size.\n");
		return -ERANGE;
	}
	ret = -EINVAL;	
	switch(pdev->type) {
	case 645:
	case 646:
		ret = set_video_mode_Nala(pdev, size, frames);
		break;

	case 675:
	case 680:
	case 690:
		ret = set_video_mode_Timon(pdev, size, frames, compression, snapshot);
		break;
		
	case 720:
	case 730:
	case 740:
	case 750:
		ret = set_video_mode_Kiara(pdev, size, frames, compression, snapshot);
		break;
	}
//	if (ret < 0) {
//		if (ret == -ENOENT)
//			Info("Video mode %s@%d fps is only supported with the decompressor module (pwcx).\n", size2name[size], frames);
//		else {
//			Err("Failed to set video mode %s@%d fps; return code = %d\n", size2name[size], frames, ret);
//		}
		return ret;
//	}
//	return 0;
}


/* BRIGHTNESS */

int wct_get_brightness(struct wct_device *pdev)
{
	char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_LUM_CTL,
		BRIGHTNESS_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return buf << 9;
}

int wct_set_brightness(struct wct_device *pdev, int value)
{
	char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	buf = (value >> 9) & 0x7f;
	return usb_control_msg(pdev->udevH, ////usb_sndctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		BRIGHTNESS_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}

/* CONTRAST */

int wct_get_contrast(struct wct_device *pdev)
{
	char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_LUM_CTL,
		CONTRAST_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return buf << 10;
}

int wct_set_contrast(struct wct_device *pdev, int value)
{
	char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	buf = (value >> 10) & 0x3f;
	return usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		CONTRAST_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}

/* GAMMA */

int wct_get_gamma(struct wct_device *pdev)
{
	char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_LUM_CTL,
		GAMMA_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return buf << 11;
}

int wct_set_gamma(struct wct_device *pdev, int value)
{
	char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	buf = (value >> 11) & 0x1f;
	return usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		GAMMA_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}


/* SATURATION */

int wct_get_saturation(struct wct_device *pdev)
{
	char buf;
	int ret;

	if (pdev->type < 675)
		return -1;
	ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_CHROM_CTL,
		pdev->type < 730 ? SATURATION_MODE_FORMATTER2 : SATURATION_MODE_FORMATTER1,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return 32768 + buf * 327;
}

int wct_set_saturation(struct wct_device *pdev, int value)
{
	char buf;

	if (pdev->type < 675)
		return -EINVAL;
	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;
	/* saturation ranges from -100 to +100 */
	buf = (value - 32768) / 327;
	return usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_CHROM_CTL,
		pdev->type < 730 ? SATURATION_MODE_FORMATTER2 : SATURATION_MODE_FORMATTER1,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}

/* AGC */

int wct_set_agc(struct wct_device *pdev, enum wct_auto_mode mode, int value)
{
	char buf;
	int ret;
	
	if (mode)
		buf = (char)0x0; /* auto */
	else
		buf = (char)0xff; /* fixed */

	ret = usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		AGC_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	
	if (!mode && ret >= 0) {
		if (value < 0)
			value = 0;
		if (value > 0xffff)
			value = 0xffff;
		buf = (value >> 10) & 0x3F;
		ret = usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
			USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			SET_LUM_CTL,
			PRESET_AGC_FORMATTER,
			pdev->vcinterface,
			&buf, 1, HZ / 2);
	}
	if (ret < 0)
		return ret;
	return 0;
}//wct_set_agc


int wct_get_agc(struct wct_device *pdev, int *value)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_LUM_CTL,
		AGC_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;

	if (buf != 0) { /* fixed */
		ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
			USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			GET_LUM_CTL,
			PRESET_AGC_FORMATTER,
			pdev->vcinterface,
			&buf, 1, HZ / 2);
		if (ret < 0)
			return ret;
		if (buf > 0x3F)
			buf = 0x3F;
		*value = (buf << 10);		
	}
	else { /* auto */
		ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
			USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			GET_STATUS_CTL,
			READ_AGC_FORMATTER,
			pdev->vcinterface,
			&buf, 1, HZ / 2);
		if (ret < 0)
			return ret;
		/* Gah... this value ranges from 0x00 ... 0x9F */
		if (buf > 0x9F)
			buf = 0x9F;
		*value = -(48 + buf * 409);
	}

	return 0;
}//wct_get_agc



int wct_set_shutter_speed(struct wct_device *pdev, enum wct_auto_mode mode, int value)
{
	char buf[2];
	int speed, ret;


	if (mode)
		buf[0] = 0x0;	/* auto */
	else
		buf[0] = (char)0xff; /* fixed */
	
	ret = usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		SHUTTER_MODE_FORMATTER,
		pdev->vcinterface,
		buf, 1, HZ / 2);

	if (!mode && ret >= 0) {
		if (value < 0)
			value = 0;
		if (value > 0xffff)
			value = 0xffff;
		switch(pdev->type) {
		case 675:
		case 680:
		case 690:
			/* speed ranges from 0x0 to 0x290 (656) */
			speed = (value / 100);
			buf[1] = speed >> 8;
			buf[0] = speed & (char)0xff;
			break;
		case 730:
		case 740:
		case 750:
			/* speed seems to range from 0x0 to 0xff */
			buf[1] = 0;
			buf[0] = value >> 8;
			break;
		}

		ret = usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
			USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			SET_LUM_CTL,
			PRESET_SHUTTER_FORMATTER,
			pdev->vcinterface,
			buf, 2, HZ / 2);
	}
	return ret;
}	


/* POWER */

int wct_camera_power(struct wct_device *pdev, int power)
{
	char buf;

	if (pdev->type < 675 || (pdev->type < 730 && pdev->release < 6))
		return 0;	/* Not supported by Nala or Timon < release 6 */

	if (power)
		buf = 0x00; /* active */
	else
		buf = (char)0xff; /* power save */
	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_STATUS_CTL,
		SET_POWER_SAVE_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}



/* private calls */

int wct_restore_user(struct wct_device *pdev)
{
	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_STATUS_CTL,
		RESTORE_USER_DEFAULTS_FORMATTER,
		pdev->vcinterface,
		NULL, 0, HZ / 2);
}

int wct_save_user(struct wct_device *pdev)
{
	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_STATUS_CTL,
		SAVE_USER_DEFAULTS_FORMATTER,
		pdev->vcinterface,
		NULL, 0, HZ / 2);
}

int wct_restore_factory(struct wct_device *pdev)
{
	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_STATUS_CTL,
		RESTORE_FACTORY_DEFAULTS_FORMATTER,
		pdev->vcinterface,
		NULL, 0, HZ / 2);
}

 /* ************************************************* */
 /* Patch by Alvarado: (not in the original version   */

 /*
  * the camera recognizes modes from 0 to 4:
  *
  * 00: indoor (incandescant lighting)
  * 01: outdoor (sunlight)
  * 02: fluorescent lighting
  * 03: manual
  * 04: auto
  */ 
int wct_set_awb(struct wct_device *pdev, enum wct_wb_mode mode)
{
	char buf;
	int ret;
	
	if (mode < WCT_WB_INDOOR)
	    mode = WCT_WB_INDOOR;
	
	if (mode > WCT_WB_AUTO)
	    mode = WCT_WB_AUTO;
	
	buf = mode & 0x07; /* just the lowest three bits */
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_CHROM_CTL,
		WB_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	
	if (ret < 0)
		return ret;
	return 0;
}//wct_set_awb

enum wct_wb_mode wct_get_awb(struct wct_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_CHROM_CTL,
		WB_MODE_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0) 
		return ret;
	return buf;
}//wct_get_awb


int wct_set_red_gain(struct wct_device *pdev, int value)
{
        unsigned char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;

	/* only the msb are considered */
	buf = value >> 8;

	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_CHROM_CTL,
		PRESET_MANUAL_RED_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}//wct_set_red_gain

int wct_get_red_gain(struct wct_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
 	        GET_CHROM_CTL, 
	        PRESET_MANUAL_RED_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
	    return ret;
	
	return (buf << 8);
}//wct_get_red_gain


int wct_set_blue_gain(struct wct_device *pdev, int value)
{
	unsigned char buf;

	if (value < 0)
		value = 0;
	if (value > 0xffff)
		value = 0xffff;

	/* linear mapping of 0..0xffff to -0x80..0x7f */
	buf = (value >> 8);

	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_CHROM_CTL,
		PRESET_MANUAL_BLUE_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}//wct_set_blue_gain

int wct_get_blue_gain(struct wct_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
   	        GET_CHROM_CTL,
		PRESET_MANUAL_BLUE_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
	    return ret;
	
	return (buf << 8);
}//wct_get_blue_gain


/* The following two functions are different, since they only read the
   internal red/blue gains, which may be different from the manual 
   gains set or read above.
 */   
int wct_read_red_gain(struct wct_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
 	        GET_STATUS_CTL, 
	        READ_RED_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
	    return ret;
	
	return (buf << 8);
}//wct_read_red_gain

int wct_read_blue_gain(struct wct_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
   	        GET_STATUS_CTL,
		READ_BLUE_GAIN_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);

	if (ret < 0)
		return ret;
	
	return (buf << 8);
}//wct_read_blue_gain


int wct_set_wb_speed(struct wct_device *pdev, int speed)
{
	unsigned char buf;
	
	/* useful range is 0x01..0x20 */
	buf = speed / 0x7f0;
	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_CHROM_CTL,
		AWB_CONTROL_SPEED_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}//wct_set_wb_speed

int wct_get_wb_speed(struct wct_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_CHROM_CTL,
		AWB_CONTROL_SPEED_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return (buf * 0x7f0);
}//wct_get_wb_speed


int wct_set_wb_delay(struct wct_device *pdev, int delay)
{
	unsigned char buf;
	
	/* useful range is 0x01..0x3F */
	buf = (delay >> 10);
	return usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_CHROM_CTL,
		AWB_CONTROL_DELAY_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}//wct_set_wb_delay

int wct_get_wb_delay(struct wct_device *pdev)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_CHROM_CTL,
		AWB_CONTROL_DELAY_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return (buf << 10);
}//wct_get_wb_delay




int wct_set_leds(struct wct_device *pdev, int on_value, int off_value)
{
	unsigned char buf[2];

	if (pdev->type < 730)
		return 0;
	on_value /= 100;
	off_value /= 100;
	if (on_value < 0)
		on_value = 0;
	if (on_value > 0xff)
		on_value = (char)0xff;
	if (off_value < 0)
		off_value = 0;
	if (off_value > 0xff)
		off_value = (char)0xff;

	buf[0] = on_value;
	buf[1] = off_value;

	return usb_control_msg(pdev->udevH, //usb_sndctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_STATUS_CTL,
		LED_FORMATTER,
		pdev->vcinterface,
		buf, 2, HZ / 2);
}//wct_set_leds

int wct_get_leds(struct wct_device *pdev, int *on_value, int *off_value)
{
	unsigned char buf[2];
	int ret;
	
	if (pdev->type < 730) {
		*on_value = -1;
		*off_value = -1;
		return 0;
	}

	ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
   	        GET_STATUS_CTL,
		LED_FORMATTER,
		pdev->vcinterface,
		buf, 2, HZ / 2);

	if (ret < 0)
		return ret;
	*on_value = buf[0] * 100;
	*off_value = buf[1] * 100;
	return 0;
}//wct_get_leds



int wct_set_contour(struct wct_device *pdev, int contour)
{
	unsigned char buf;
	int ret;
	
	if (contour < 0)
		buf = (char)0xff; /* auto contour on */
	else
		buf = 0x0; /* auto contour off */
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		AUTO_CONTOUR_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	
	if (contour < 0)
		return 0;
	if (contour > 0xffff)
		contour = 0xffff;
	
	buf = (contour >> 10); /* contour preset is [0..3f] */
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		PRESET_CONTOUR_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)	
		return ret;	
	return 0;
}//wct_set_contour

int wct_get_contour(struct wct_device *pdev, int *contour)
{
	unsigned char buf;
	int ret;
	
	ret = usb_control_msg(pdev->udevH,
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_LUM_CTL,
		AUTO_CONTOUR_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;

	if (buf == 0) {
		/* auto mode off, query current preset value */
		ret = usb_control_msg(pdev->udevH,
			USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			GET_LUM_CTL,
			PRESET_CONTOUR_FORMATTER,
			pdev->vcinterface,
			&buf, 1, HZ / 2);
		if (ret < 0)	
			return ret;
		*contour =  (buf << 10);
	}
	else
		*contour = -1;
	return 0;
}//wct_get_contour



int wct_set_backlight(struct wct_device *pdev, enum wct_onoff_mode backlight)
{
	unsigned char buf;
	
	buf = (char)backlight;

	return usb_control_msg(pdev->udevH, 
		USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		SET_LUM_CTL,
		BACK_LIGHT_COMPENSATION_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
}//wct_set_backlight

enum wct_onoff_mode wct_get_backlight(struct wct_device *pdev)
{
	int ret;
	unsigned char buf;
	
	ret = usb_control_msg(pdev->udevH, 
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_LUM_CTL,
		BACK_LIGHT_COMPENSATION_FORMATTER,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	return buf;
}//wct_get_backlight



int wct_set_flicker(struct wct_device *pdev, enum wct_onoff_mode flicker)
{
	unsigned char buf;
	
	if (flicker)
		buf = (char)0xff;
	else
		buf = 0x0;
	return SendControlMsg(SET_LUM_CTL, FLICKERLESS_MODE_FORMATTER, 1);
}//wct_set_flicker

enum wct_onoff_mode wct_get_flicker(struct wct_device *pdev)
{
	int ret;
	unsigned char buf;
	
	ret = RecvControlMsg(GET_LUM_CTL, FLICKERLESS_MODE_FORMATTER, 1);
	if (ret < 0)
		return ret;
	return buf;
}//wct_get_flicker



int wct_set_dynamic_noise(struct wct_device *pdev, enum wct_dynnoise_mode noise)
{
	unsigned char buf;

	if (noise < WCT_DYNN_OFF)
		noise = WCT_DYNN_OFF;
	if (noise > WCT_DYYN_FULL)
		noise = WCT_DYYN_FULL;
	buf = noise;
	return SendControlMsg(SET_LUM_CTL, DYNAMIC_NOISE_CONTROL_FORMATTER, 1);
}//wct_set_dynamic_noise

enum wct_dynnoise_mode wct_get_dynamic_noise(struct wct_device *pdev)
{
	int ret;
	unsigned char buf;
	
	ret = RecvControlMsg(GET_LUM_CTL, DYNAMIC_NOISE_CONTROL_FORMATTER, 1);
	if (ret < 0)
		return ret;
	return buf;
}


int wct_get_cmos_sensor(struct wct_device *pdev)
{
	unsigned char buf;
	int ret = -1, request;
	
	if (pdev->type < 675)
		request = SENSOR_TYPE_FORMATTER1;
	else if (pdev->type < 730)
		return -1; /* The Vesta series doesn't have this call */
	else
		request = SENSOR_TYPE_FORMATTER2;
	
	ret = usb_control_msg(pdev->udevH, //usb_rcvctrlpipe(pdev->udevH, 0),
		USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
		GET_STATUS_CTL,
		request,
		pdev->vcinterface,
		&buf, 1, HZ / 2);
	if (ret < 0)
		return ret;
	if (pdev->type < 675)
		return buf | 0x100;
	else
		return buf;
}//wct_get_cmos_sensor

int wct_set_colourmode(struct wct_device *pdev, enum wct_onoff_mode colourmode) {
         unsigned char buf;

         buf = (char)colourmode;

         return usb_control_msg(pdev->udevH,
                 USB_ENDPOINT_OUT | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                 SET_CHROM_CTL,
                 COLOUR_MODE_FORMATTER,
                 pdev->vcinterface,
                 &buf, 1, HZ / 2);
}//wct_set_colourmode

enum wct_onoff_mode wct_get_colourmode(struct wct_device *pdev) {
         int ret;
         unsigned char buf;

         ret = usb_control_msg(pdev->udevH,
                 USB_ENDPOINT_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
                 SET_CHROM_CTL,
                 COLOUR_MODE_FORMATTER,
                 pdev->vcinterface,
                 &buf, 1, HZ / 2);
         if (ret < 0)
                 return ret;
         return buf;
}//wct_get_colourmode
 

